"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const minimist_1 = (0, tslib_1.__importDefault)(require("minimist"));
const typescript_1 = (0, tslib_1.__importDefault)(require("typescript"));
const fs = (0, tslib_1.__importStar)(require("fs"));
const path = (0, tslib_1.__importStar)(require("path"));
const chokidar = (0, tslib_1.__importStar)(require("chokidar"));
const fs_extra_1 = (0, tslib_1.__importDefault)(require("fs-extra"));
const fast_glob_1 = (0, tslib_1.__importDefault)(require("fast-glob"));
const core_1 = require("./core");
const typescriptGenerator = (0, tslib_1.__importStar)(require("./typescript-generator"));
const packageJson = (0, tslib_1.__importStar)(require("../package.json"));
const protobuf = (0, tslib_1.__importStar)(require("protobufjs"));
const ajv_1 = (0, tslib_1.__importDefault)(require("ajv"));
const ajv = new ajv_1.default();
async function executeCommandLine() {
    const argv = (0, minimist_1.default)(process.argv.slice(2), { '--': true });
    const showVersion = argv.v || argv.version;
    if (showVersion) {
        showToolVersion();
        return;
    }
    if (argv.h || argv.help) {
        showHelp();
        return;
    }
    const { protobufPath, graphqlPath, graphqlRootTypePath, reasonPath, ocamlPath, rustPath, jsonPath, mongoosePath, swaggerPath, swaggerBasePath, debugPath, filePaths, watchMode, looseMode, customPath, configPaths, typescriptPath, markdownPath, } = parseParameters(argv);
    function generateJsonSchemas(generator) {
        if (!jsonPath) {
            return;
        }
        fs_extra_1.default.ensureDirSync(jsonPath);
        const schemas = generator.generateJsonSchemas();
        for (const { entry, schema } of schemas) {
            if ((debugPath || ajv.validateSchema(schema)) && entry) {
                fs.writeFileSync(path.resolve(jsonPath, entry), JSON.stringify(schema, null, '  '));
            }
            else {
                printInConsole(`json schema verified fail for entry: ${entry}`);
                if (!watchMode) {
                    process.exit(1);
                }
            }
        }
    }
    let program;
    async function run() {
        const files = await (0, fast_glob_1.default)(filePaths);
        const newProgram = typescript_1.default.createProgram(files, { target: typescript_1.default.ScriptTarget.ESNext }, undefined, program);
        program = newProgram;
        const sourceFiles = files.map(filePath => newProgram.getSourceFile(filePath));
        const generator = new core_1.Generator(sourceFiles.filter((s) => !!s), looseMode, !!configPaths, fileName => path.relative(process.cwd(), fileName), program.getTypeChecker());
        if (debugPath) {
            fs.writeFileSync(debugPath, JSON.stringify(generator.declarations, null, '  '));
        }
        if (protobufPath) {
            const protobufContent = `/**
 * This file is generated by 'types-as-schema'
 * It is not mean to be edited by hand
 */
` + generator.generateProtobuf();
            protobuf.parse(protobufContent);
            fs.writeFileSync(protobufPath, protobufContent);
        }
        if (graphqlPath) {
            const graphqlContent = generator.generateGraphqlSchema();
            fs.writeFileSync(graphqlPath, graphqlContent);
        }
        if (graphqlRootTypePath) {
            const graphqlRootPathContent = generator.generateGraphqlRootType((referenceTypes) => getReferenceTypeImports(referenceTypes, graphqlRootTypePath));
            fs.writeFileSync(graphqlRootTypePath, graphqlRootPathContent);
        }
        if (reasonPath) {
            const reasonContent = generator.generateReasonTypes();
            fs.writeFileSync(reasonPath, reasonContent);
        }
        if (ocamlPath) {
            const ocamlContent = generator.generateOcamlTypes();
            fs.writeFileSync(ocamlPath, ocamlContent);
        }
        if (rustPath) {
            const rustContent = generator.generateRustTypes();
            fs.writeFileSync(rustPath, rustContent);
        }
        if (mongoosePath) {
            const mongooseContent = generator.generateMongooseSchema();
            fs.writeFileSync(mongoosePath, mongooseContent);
        }
        if (swaggerPath) {
            const swaggerBase = swaggerBasePath ? JSON.parse(fs.readFileSync(swaggerBasePath).toString()) : undefined;
            const swaggerDoc = generator.generateSwaggerDoc(swaggerBase);
            fs.writeFileSync(swaggerPath, swaggerDoc);
        }
        if (jsonPath) {
            generateJsonSchemas(generator);
        }
        if (configPaths) {
            for (const configPath of configPaths) {
                let configFilePath;
                if (path.isAbsolute(configPath)) {
                    configFilePath = configPath;
                }
                else if (configPath.startsWith(`.${path.sep}`)
                    || configPath.startsWith(`..${path.sep}`)
                    || configPath.startsWith('./')
                    || configPath.startsWith('../')) {
                    configFilePath = path.resolve(process.cwd(), configPath);
                }
                else {
                    configFilePath = configPath;
                }
                if (configFilePath.endsWith('.ts')) {
                    require('ts-node/register/transpile-only');
                }
                // eslint-disable-next-line @typescript-eslint/no-var-requires
                let action = require(configFilePath);
                if (action.default) {
                    action = action.default;
                }
                const customContent = action(generator.declarations, typescriptGenerator);
                if (typeof customContent === 'string' && customPath) {
                    fs.writeFileSync(customPath, customContent);
                }
                else if (Array.isArray(customContent)) {
                    for (const r of customContent) {
                        fs.writeFileSync(r.path, r.content);
                    }
                }
            }
        }
        if (typescriptPath) {
            const typescriptContent = generator.generateTypescript();
            fs.writeFileSync(typescriptPath, typescriptContent);
        }
        if (markdownPath) {
            const markdownContent = generator.generateMarkdownDoc();
            fs.writeFileSync(markdownPath, markdownContent);
        }
    }
    if (watchMode) {
        let ready = false;
        chokidar.watch(filePaths).on('ready', () => {
            ready = true;
            run();
        }).on('all', (type, file) => {
            printInConsole(`Detecting ${type}: ${file}`);
            if (ready) {
                run();
            }
        });
    }
    else {
        await run();
    }
}
function parseParameter(argv, name) {
    const result = argv[name];
    if (result && typeof result !== 'string') {
        throw new Error(`expect the path of generated ${name} file`);
    }
    return result;
}
function parseConfigPaths(argv) {
    const result = argv.config;
    if (!result) {
        return undefined;
    }
    if (typeof result === 'string') {
        return [result];
    }
    if (Array.isArray(result) && result.length > 0) {
        return result;
    }
    throw new Error(`expect the path of generated config file`);
}
function parseParameters(argv) {
    const protobufPath = parseParameter(argv, 'protobuf');
    const graphqlPath = parseParameter(argv, 'graphql');
    const graphqlRootTypePath = parseParameter(argv, 'graphql-root-type');
    const reasonPath = parseParameter(argv, 'reason');
    const ocamlPath = parseParameter(argv, 'ocaml');
    const rustPath = parseParameter(argv, 'rust');
    const jsonPath = parseParameter(argv, 'json');
    const mongoosePath = parseParameter(argv, 'mongoose');
    const swaggerPath = parseParameter(argv, 'swagger');
    const swaggerBasePath = parseParameter(argv, 'swagger-base');
    const debugPath = parseParameter(argv, 'debug');
    const customPath = parseParameter(argv, 'custom');
    const configPaths = parseConfigPaths(argv);
    const typescriptPath = parseParameter(argv, 'typescript');
    const markdownPath = parseParameter(argv, 'markdown');
    const filePaths = argv._;
    if (filePaths.length === 0) {
        throw new Error('expect the path of types file');
    }
    const watchMode = argv.w || argv.watch;
    const looseMode = argv.loose;
    return {
        protobufPath,
        graphqlPath,
        graphqlRootTypePath,
        reasonPath,
        ocamlPath,
        rustPath,
        jsonPath,
        mongoosePath,
        swaggerPath,
        swaggerBasePath,
        debugPath,
        filePaths,
        watchMode,
        looseMode,
        customPath,
        configPaths,
        typescriptPath,
        markdownPath,
    };
}
function printInConsole(message) {
    console.log(message);
}
function showToolVersion() {
    printInConsole(`Version: ${packageJson.version}`);
}
function showHelp() {
    console.log(`Version ${packageJson.version}
Syntax:   types-as-schema [options] [file...]
Examples: types-as-schema demo/types.ts --protobuf demo/types.proto
Options:
 -h, --help                                         Print this message.
 -v, --version                                      Print the version
 --json                                             directory for generated json files
 --protobuf                                         generated protobuf file
 --graphql                                          generated graphql schema file
 --graphql-root-type                                generated graphql root type
 --reason                                           generated reason types file
 --ocaml                                            generated ocaml types file
 --rust                                             generated rust types file
 --mongoose                                         generated mongoose schema file
 --swagger                                          generated swagger json file
 --swagger-base                                     swagger json file that generation based on
 --typescript                                       generated typescript file
 --debug                                            generated file with debug information in it
 --watch, -w                                        watch mode
 --loose                                            do not force additionalProperties
 --config                                           generate file by the config file, can be multiple
 --markdown                                         generated markdown file
`);
}
executeCommandLine().then(() => {
    printInConsole('types-as-schema success.');
}, (error) => {
    printInConsole(error);
    process.exit(1);
});
function getReferenceTypeImports(referenceTypes, graphqlRootTypePath) {
    const map = {};
    for (const referenceType of referenceTypes) {
        const file = referenceType.position.file;
        if (!map[file]) {
            map[file] = [];
        }
        const references = map[file];
        if (references?.every((n) => n !== referenceType.name)) {
            references.push(referenceType.name);
        }
    }
    const dirname = path.dirname(graphqlRootTypePath);
    const imports = [];
    for (const file in map) {
        let relativePath = path.relative(dirname, file);
        if (!relativePath.startsWith('.' + path.sep) && !relativePath.startsWith('..' + path.sep)) {
            relativePath = '.' + path.sep + relativePath;
        }
        relativePath = relativePath.substring(0, relativePath.length - path.extname(relativePath).length);
        imports.push(`import { ${map[file]?.join(', ')} } from '${relativePath}'`);
    }
    return imports.join('\n');
}
