"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Parser = void 0;
const tslib_1 = require("tslib");
const typescript_1 = (0, tslib_1.__importDefault)(require("typescript"));
const js_doc_1 = require("ts-lib-utils/dist/js-doc");
const utils_1 = require("./utils");
class Parser {
    sourceFiles;
    getRelativePath;
    checker;
    declarations = [];
    disableWarning = false;
    constructor(sourceFiles, getRelativePath, checker) {
        this.sourceFiles = sourceFiles;
        this.getRelativePath = getRelativePath;
        this.checker = checker;
    }
    parse() {
        for (const sourceFile of this.sourceFiles) {
            typescript_1.default.forEachChild(sourceFile, node => {
                if (typescript_1.default.isEnumDeclaration(node)) {
                    this.preHandleEnumDeclaration(node, sourceFile);
                }
            });
        }
        for (const sourceFile of this.sourceFiles) {
            typescript_1.default.forEachChild(sourceFile, node => {
                this.handleSourceFile(node, sourceFile);
            });
        }
        return this.declarations;
    }
    preHandleEnumDeclaration(declaration, sourceFile) {
        const members = declaration.members;
        if (members.length > 0) {
            const firstMember = members[0];
            if (firstMember?.initializer) {
                this.handleEnumDeclarationInitializer(declaration, members, firstMember.initializer, sourceFile);
            }
            else {
                const enumType = {
                    kind: 'enum',
                    name: declaration.name.text,
                    type: 'uint32',
                    members: [],
                    position: this.getPosition(declaration, sourceFile),
                    modifiers: this.getModifiers(declaration),
                };
                let lastIndex = 0;
                for (const member of members) {
                    if (member.initializer && typescript_1.default.isIdentifier(member.name)) {
                        const { name, value } = this.getExpression(member.initializer, member.name, sourceFile);
                        enumType.members.push({
                            name,
                            value: value
                        });
                        lastIndex = value + 1;
                    }
                    else {
                        enumType.members.push({
                            name: member.name.text,
                            value: lastIndex
                        });
                        lastIndex++;
                    }
                }
                this.declarations.push(enumType);
            }
        }
    }
    handleEnumDeclarationInitializer(declaration, members, initializer, sourceFile) {
        const enumType = {
            kind: 'enum',
            name: declaration.name.text,
            type: typescript_1.default.isStringLiteral(initializer) ? 'string' : 'uint32',
            members: [],
            position: this.getPosition(declaration, sourceFile),
            modifiers: this.getModifiers(declaration),
        };
        for (const member of members) {
            if (member.initializer && typescript_1.default.isIdentifier(member.name)) {
                const { name, value } = this.getExpression(member.initializer, member.name, sourceFile);
                enumType.members.push({
                    name,
                    value: value
                });
            }
        }
        this.declarations.push(enumType);
    }
    handleSourceFile(node, sourceFile) {
        const jsDocs = this.getJsDocs(node, sourceFile);
        if (typescript_1.default.isTypeAliasDeclaration(node)) {
            this.handleTypeAliasDeclaration(node, jsDocs, sourceFile);
        }
        else if (typescript_1.default.isInterfaceDeclaration(node) || typescript_1.default.isClassDeclaration(node)) {
            this.handleInterfaceOrClassDeclaration(node, jsDocs, sourceFile);
        }
        else if (typescript_1.default.isFunctionDeclaration(node)) {
            this.handleFunctionDeclaration(node, jsDocs, sourceFile);
        }
        else if (typescript_1.default.isExportAssignment(node)) {
            if (typescript_1.default.isArrowFunction(node.expression)) {
                this.handleFunctionDeclaration(node.expression, jsDocs, sourceFile);
            }
        }
        else if (typescript_1.default.isVariableStatement(node)) {
            const declaration = node.declarationList.declarations[0];
            if (declaration?.initializer && typescript_1.default.isArrowFunction(declaration.initializer)) {
                this.handleFunctionDeclaration(declaration.initializer, jsDocs, sourceFile, typescript_1.default.isIdentifier(declaration.name) ? declaration.name.text : undefined);
            }
        }
    }
    handleFunctionDeclaration(declaration, { jsDocs, comments }, sourceFile, functionName) {
        const type = declaration.type
            ? this.getType(declaration.type, sourceFile)
            : {
                kind: undefined,
                position: this.getPosition(declaration, sourceFile)
            };
        const functionDeclaration = {
            kind: 'function',
            name: functionName ?? (declaration.name && !typescript_1.default.isFunctionTypeNode(declaration) ? declaration.name.text : ''),
            type,
            optional: typescript_1.default.isFunctionTypeNode(declaration) ? false : !!declaration.questionToken,
            parameters: declaration.parameters.map((parameter) => this.handleFunctionParameter(parameter, sourceFile)),
            comments,
            jsDocs,
            position: this.getPosition(declaration, sourceFile),
            body: typescript_1.default.isFunctionTypeNode(declaration) ? undefined : declaration.body?.getText(sourceFile),
            modifiers: this.getModifiers(declaration),
        };
        for (const jsDoc of jsDocs || []) {
            if (jsDoc.comment) {
                if (jsDoc.name === 'method') {
                    functionDeclaration.method = jsDoc.comment;
                }
                else if (jsDoc.name === 'path') {
                    functionDeclaration.path = jsDoc.comment;
                }
                else if (jsDoc.name === 'description') {
                    functionDeclaration.description = jsDoc.comment;
                }
                else if (jsDoc.name === 'summary') {
                    functionDeclaration.summary = jsDoc.comment;
                }
                else if (jsDoc.name === 'tags') {
                    functionDeclaration.tags = jsDoc.comment.split(',');
                }
            }
            else if (jsDoc.name === 'deprecated') {
                functionDeclaration.deprecated = true;
            }
        }
        this.declarations.push(functionDeclaration);
    }
    handleFunctionParameter(parameter, sourceFile) {
        const parameterDoc = this.getParameter(parameter, sourceFile);
        const { jsDocs, comments } = this.getJsDocs(parameter, sourceFile);
        parameterDoc.comments = comments;
        parameterDoc.jsDocs = jsDocs;
        for (const jsDoc of jsDocs || []) {
            if (jsDoc.comment && jsDoc.name === 'in') {
                parameterDoc.in = jsDoc.comment;
            }
            else {
                this.setJsDoc(jsDoc, parameterDoc.type);
            }
        }
        if (parameterDoc.type.default !== undefined) {
            parameterDoc.optional = true;
        }
        return parameterDoc;
    }
    handleInterfaceOrClassDeclaration(declaration, { jsDocs, comments }, sourceFile) {
        if (declaration.name) {
            const declarationName = declaration.name.text;
            // if the node is pre-handled, then it should be in `declarations` already, so don't continue
            if (this.declarations.some(m => m.kind === 'reference' ? m.newName === declarationName : m.name === declarationName)) {
                return;
            }
        }
        const objectMembers = this.getObjectMembers(declaration.members, sourceFile);
        const members = objectMembers.members;
        let minProperties = objectMembers.minProperties;
        let maxProperties = objectMembers.maxProperties;
        let additionalProperties = objectMembers.additionalProperties;
        const heritageClauses = this.handleHeritageClauses(declaration, members, minProperties, maxProperties);
        minProperties = heritageClauses.minProperties;
        maxProperties = heritageClauses.maxProperties;
        if (heritageClauses.additionalProperties) {
            additionalProperties = heritageClauses.additionalProperties;
        }
        const objectDeclaration = {
            kind: 'object',
            name: declaration.name ? declaration.name.text : '',
            members,
            minProperties,
            maxProperties: additionalProperties === undefined ? maxProperties : undefined,
            additionalProperties,
            entry: jsDocs?.find((f) => f.name === 'entry')?.comment,
            position: this.getPosition(declaration, sourceFile),
            comments,
            jsDocs,
            decorators: typescript_1.default.isClassDeclaration(declaration) ? this.getDecorators(sourceFile, declaration.decorators) : undefined,
            modifiers: this.getModifiers(declaration),
        };
        for (const jsDoc of jsDocs || []) {
            this.setJsDocObject(jsDoc, objectDeclaration);
        }
        this.declarations.push(objectDeclaration);
    }
    handleHeritageClauses(declaration, members, minProperties, maxProperties) {
        let additionalProperties;
        if (declaration.heritageClauses) {
            for (const clause of declaration.heritageClauses) {
                if (typescript_1.default.isHeritageClause(clause)) {
                    for (const type of clause.types) {
                        if (typescript_1.default.isExpressionWithTypeArguments(type) && typescript_1.default.isIdentifier(type.expression)) {
                            ({ minProperties, maxProperties, additionalProperties } = this.handleExpressionWithTypeArguments(type.expression, members, minProperties, maxProperties));
                        }
                    }
                }
            }
        }
        return { minProperties, maxProperties, additionalProperties };
    }
    handleExpressionWithTypeArguments(declaration, members, minProperties, maxProperties) {
        const interfaceName = declaration.text;
        this.preHandleType(interfaceName);
        let additionalProperties;
        const clauseDeclaration = this.declarations.find(m => m.kind === 'object' && m.name === interfaceName);
        if (clauseDeclaration && clauseDeclaration.kind === 'object') {
            additionalProperties = clauseDeclaration.additionalProperties;
            for (const member of clauseDeclaration.members) {
                if (members.every(m => m.name !== member.name)) {
                    members.push(member);
                    maxProperties++;
                    if (!member.optional) {
                        minProperties++;
                    }
                }
            }
        }
        return { minProperties, maxProperties, additionalProperties };
    }
    handleTypeAliasDeclaration(declaration, jsDocs, sourceFile) {
        if (typescript_1.default.isArrayTypeNode(declaration.type)) {
            this.handleArrayTypeInTypeAliasDeclaration(declaration.type, declaration.name, jsDocs, sourceFile);
        }
        else if (typescript_1.default.isTypeLiteralNode(declaration.type)
            || typescript_1.default.isUnionTypeNode(declaration.type)
            || typescript_1.default.isIntersectionTypeNode(declaration.type)) {
            this.handleTypeLiteralOrUnionTypeOrIntersectionType(declaration.type, declaration.name, jsDocs, sourceFile);
        }
        else if (typescript_1.default.isTypeReferenceNode(declaration.type) &&
            typescript_1.default.isIdentifier(declaration.type.typeName) &&
            declaration.type.typeArguments &&
            declaration.type.typeArguments.length > 0) {
            if (this.checker) {
                const type = this.getTypeOfComplexType(declaration.type, sourceFile);
                if (type) {
                    this.declarations.push({
                        name: declaration.name.text,
                        ...type,
                        modifiers: this.getModifiers(declaration),
                    });
                    return;
                }
                else if (!this.disableWarning) {
                    (0, utils_1.warn)(this.getPosition(declaration, sourceFile), 'parse');
                }
            }
            else if (declaration.type.typeName.text === 'Pick') {
                const type = this.getTypeOfPick(declaration.type.typeName, declaration.type.typeArguments, sourceFile);
                if (type) {
                    this.declarations.push({
                        name: declaration.name.text,
                        ...type,
                        modifiers: this.getModifiers(declaration),
                    });
                    return;
                }
            }
            else if (!this.disableWarning) {
                (0, utils_1.warn)(this.getPosition(declaration, sourceFile), 'parse');
            }
        }
        else if (typescript_1.default.isTemplateLiteralTypeNode(declaration.type)) {
            const spans = this.getTemplateSpans(declaration.type, sourceFile);
            if (spans.every((s) => s.kind === 'enum')) {
                this.declarations.push({
                    kind: 'string',
                    name: declaration.name.text,
                    enums: this.getTemplateLiteralTypeEnums(declaration.type, spans),
                    position: this.getPosition(declaration.name, sourceFile),
                    modifiers: this.getModifiers(declaration),
                });
            }
            else if (spans.every((s) => s.kind === 'enum' || s.kind === 'string' || s.kind === 'number' || s.kind === 'boolean')) {
                this.declarations.push({
                    kind: 'string',
                    name: declaration.name.text,
                    templateLiteral: this.getTemplateLiteral(declaration.type, spans),
                    position: this.getPosition(declaration.type, sourceFile),
                    modifiers: this.getModifiers(declaration),
                });
            }
        }
        else if (typescript_1.default.isFunctionTypeNode(declaration.type)) {
            this.handleFunctionDeclaration(declaration.type, jsDocs, sourceFile, typescript_1.default.isIdentifier(declaration.name) ? declaration.name.text : undefined);
        }
        else {
            const type = this.getType(declaration.type, sourceFile);
            const declarationName = declaration.name && !typescript_1.default.isFunctionTypeNode(declaration) ? declaration.name.text : '';
            if (type.kind === 'reference') {
                this.declarations.push({
                    newName: declarationName,
                    ...type,
                    ...jsDocs,
                    modifiers: this.getModifiers(declaration),
                });
            }
            else if (type.kind === 'enum') {
                this.declarations.push({
                    members: type.enums.map((e) => ({
                        name: e,
                        value: e,
                    })),
                    ...type,
                    ...jsDocs,
                    name: declarationName,
                    modifiers: this.getModifiers(declaration),
                });
            }
            else if (type.kind !== undefined &&
                type.kind !== 'file' &&
                type.kind !== 'void') {
                this.declarations.push({
                    ...type,
                    ...jsDocs,
                    name: declarationName,
                    modifiers: this.getModifiers(declaration),
                });
            }
        }
    }
    handleTypeLiteralOrUnionTypeOrIntersectionType(declarationType, declarationName, { jsDocs, comments }, sourceFile) {
        let unionDeclaration;
        if (typescript_1.default.isUnionTypeNode(declarationType)) {
            if (declarationType.types.every(u => typescript_1.default.isLiteralTypeNode(u) || u.kind === typescript_1.default.SyntaxKind.NullKeyword)) {
                this.handleUnionTypeOfLiteralType(declarationType, declarationName, sourceFile);
                return;
            }
            if (declarationType.types.every(u => typescript_1.default.isTypeReferenceNode(u))) {
                unionDeclaration = {
                    kind: 'union',
                    name: declarationName.text,
                    members: declarationType.types.map(u => this.getType(u, sourceFile)),
                    entry: jsDocs?.find((f) => f.name === 'entry')?.comment,
                    position: this.getPosition(declarationName, sourceFile),
                    comments,
                    jsDocs,
                    modifiers: this.getModifiers(declarationType),
                };
                this.declarations.push(unionDeclaration);
            }
        }
        const { members, minProperties, maxProperties, additionalProperties } = this.getMembersInfo(declarationType, sourceFile);
        const objectDeclaration = {
            kind: 'object',
            name: declarationName.text,
            members,
            minProperties,
            maxProperties: additionalProperties === undefined ? maxProperties : undefined,
            additionalProperties,
            entry: jsDocs?.find((f) => f.name === 'entry')?.comment,
            position: this.getPosition(declarationName, sourceFile),
            comments,
            jsDocs,
            modifiers: this.getModifiers(declarationType),
        };
        for (const jsDoc of jsDocs || []) {
            this.setJsDocObject(jsDoc, objectDeclaration);
        }
        if (unionDeclaration) {
            unionDeclaration.objectType = objectDeclaration;
        }
        else {
            this.declarations.push(objectDeclaration);
        }
    }
    handleUnionTypeOfLiteralType(unionType, declarationName, sourceFile) {
        let enumType;
        const enums = [];
        for (const childType of unionType.types) {
            if (typescript_1.default.isLiteralTypeNode(childType)) {
                const { type, value } = this.getEnumOfLiteralType(childType);
                if (type !== undefined) {
                    enumType = type;
                }
                if (value !== undefined) {
                    enums.push(value);
                }
            }
            else if (childType.kind === typescript_1.default.SyntaxKind.NullKeyword) {
                enums.push(null);
            }
        }
        if (enumType) {
            if (enumType === 'string') {
                const stringDeclaration = {
                    kind: 'string',
                    name: declarationName.text,
                    enums: enums,
                    position: this.getPosition(declarationName, sourceFile),
                    modifiers: this.getModifiers(unionType),
                };
                this.declarations.push(stringDeclaration);
            }
            else if (enumType === 'number') {
                const numberDeclaration = {
                    kind: 'number',
                    type: enumType,
                    name: declarationName.text,
                    enums: enums,
                    position: this.getPosition(declarationName, sourceFile),
                    modifiers: this.getModifiers(unionType),
                };
                this.declarations.push(numberDeclaration);
            }
            else if (enumType === 'boolean') {
                const unionDeclaration = {
                    kind: 'union',
                    name: declarationName.text,
                    members: unionType.types.map(e => this.getType(e, sourceFile)),
                    entry: undefined,
                    position: this.getPosition(declarationName, sourceFile),
                    modifiers: this.getModifiers(unionType),
                };
                this.declarations.push(unionDeclaration);
            }
        }
    }
    handleArrayTypeInTypeAliasDeclaration(arrayType, declarationName, { jsDocs, comments }, sourceFile) {
        const type = this.getType(arrayType.elementType, sourceFile);
        const arrayDeclaration = {
            kind: 'array',
            name: declarationName.text,
            type,
            entry: jsDocs?.find((f) => f.name === 'entry')?.comment,
            position: this.getPosition(declarationName, sourceFile),
            comments,
            jsDocs,
            modifiers: this.getModifiers(arrayType),
        };
        for (const jsDoc of jsDocs || []) {
            this.setJsDocArray(jsDoc, arrayDeclaration);
        }
        this.declarations.push(arrayDeclaration);
    }
    getJsDocs(node, sourceFile) {
        const ranges = typescript_1.default.getLeadingCommentRanges(sourceFile.text, node.getFullStart());
        const comments = ranges ? ranges.map((r) => sourceFile.text.slice(r.pos, r.end)) : undefined;
        const jsDocs = (0, js_doc_1.getJsDocs)(node);
        const result = [];
        for (const jsDoc of jsDocs) {
            result.push({
                ...jsDoc,
                type: jsDoc.type ? this.getType(jsDoc.type, sourceFile) : undefined
            });
        }
        return { jsDocs: result.length === 0 ? undefined : result, comments };
    }
    getType(type, sourceFile) {
        if (type.kind === typescript_1.default.SyntaxKind.StringKeyword) {
            return {
                kind: 'string',
                position: this.getPosition(type, sourceFile)
            };
        }
        if (type.kind === typescript_1.default.SyntaxKind.NumberKeyword) {
            return {
                kind: 'number',
                type: 'number',
                position: this.getPosition(type, sourceFile)
            };
        }
        if (type.kind === typescript_1.default.SyntaxKind.BooleanKeyword) {
            return {
                kind: 'boolean',
                position: this.getPosition(type, sourceFile)
            };
        }
        if (type.kind === typescript_1.default.SyntaxKind.VoidKeyword) {
            return {
                kind: 'void',
                position: this.getPosition(type, sourceFile)
            };
        }
        if (typescript_1.default.isTypeLiteralNode(type)) {
            return this.getTypeOfTypeLiteral(type, sourceFile);
        }
        if (typescript_1.default.isArrayTypeNode(type)) {
            const elementType = this.getType(type.elementType, sourceFile);
            return {
                kind: 'array',
                type: elementType,
                position: this.getPosition(type, sourceFile)
            };
        }
        if (typescript_1.default.isTypeReferenceNode(type)) {
            return this.getTypeOfTypeReference(type, sourceFile);
        }
        if (typescript_1.default.isUnionTypeNode(type) || typescript_1.default.isIntersectionTypeNode(type)) {
            return this.getTypeOfUnionTypeOrIntersectionType(type, sourceFile);
        }
        if (typescript_1.default.isLiteralTypeNode(type)) {
            if (type.literal.kind === typescript_1.default.SyntaxKind.NullKeyword) {
                return {
                    kind: 'null',
                    position: this.getPosition(type, sourceFile)
                };
            }
            return this.getTypeOfLiteralType(type, sourceFile);
        }
        if (type.kind === typescript_1.default.SyntaxKind.NullKeyword) {
            return {
                kind: 'null',
                position: this.getPosition(type, sourceFile)
            };
        }
        if (typescript_1.default.isTupleTypeNode(type)) {
            let arrayType;
            const elementsType = type.elements || type.elementTypes;
            for (const elementType of elementsType) {
                arrayType = this.getType(elementType, sourceFile);
            }
            if (arrayType) {
                return {
                    kind: 'array',
                    type: arrayType,
                    minItems: elementsType.length,
                    maxItems: elementsType.length,
                    position: this.getPosition(type, sourceFile)
                };
            }
        }
        if (typescript_1.default.isTemplateLiteralTypeNode(type)) {
            const spans = this.getTemplateSpans(type, sourceFile);
            if (spans.every((s) => s.kind === 'enum')) {
                return {
                    kind: 'enum',
                    type: 'string',
                    name: 'string',
                    enums: this.getTemplateLiteralTypeEnums(type, spans),
                    position: this.getPosition(type, sourceFile)
                };
            }
            else if (spans.every((s) => s.kind === 'enum' || s.kind === 'string' || s.kind === 'number' || s.kind === 'boolean')) {
                return {
                    kind: 'string',
                    templateLiteral: this.getTemplateLiteral(type, spans),
                    position: this.getPosition(type, sourceFile)
                };
            }
        }
        if (typescript_1.default.isTemplateMiddleOrTemplateTail(type)) {
            return {
                kind: 'enum',
                type: 'string',
                name: 'string',
                enums: [type.text],
                position: this.getPosition(type, sourceFile),
            };
        }
        if (typescript_1.default.isParenthesizedTypeNode(type)) {
            return this.getType(type.type, sourceFile);
        }
        if (typescript_1.default.isTypeOperatorNode(type)) {
            if (type.operator === typescript_1.default.SyntaxKind.KeyOfKeyword && typescript_1.default.isTypeReferenceNode(type.type)) {
                const objectType = this.getTypeOfComplexType(type.type, sourceFile);
                if (objectType) {
                    return {
                        kind: 'enum',
                        type: 'string',
                        name: 'string',
                        enums: objectType.members.map((m) => m.name),
                        position: this.getPosition(type, sourceFile),
                    };
                }
            }
        }
        const position = this.getPosition(type, sourceFile);
        if (type.kind !== typescript_1.default.SyntaxKind.AnyKeyword && !this.disableWarning) {
            (0, utils_1.warn)(position, 'parser');
        }
        return {
            kind: undefined,
            position
        };
    }
    getTemplateSpans(type, sourceFile) {
        const spans = [];
        for (const span of type.templateSpans) {
            spans.push(this.getType(span.type, sourceFile), this.getType(span.literal, sourceFile));
        }
        return spans;
    }
    getTemplateLiteralTypeEnums(type, spans) {
        let parts = [type.head.text];
        for (const t of spans) {
            if (t.kind === 'enum') {
                const newParts = [];
                for (const e of t.enums) {
                    newParts.push(...parts.map((p) => p + e));
                }
                parts = newParts;
            }
        }
        return parts;
    }
    getTemplateLiteral(type, spans) {
        const templateLiteral = [];
        if (type.head.text) {
            templateLiteral.push({
                kind: 'enum',
                enums: [type.head.text],
            });
        }
        for (const t of spans) {
            if (t.kind === 'string') {
                templateLiteral.push({
                    kind: 'string'
                });
            }
            else if (t.kind === 'number') {
                templateLiteral.push({
                    kind: 'number'
                });
            }
            else if (t.kind === 'boolean') {
                templateLiteral.push({
                    kind: 'boolean'
                });
            }
            else if (t.kind === 'enum') {
                const enums = t.enums.filter((e) => e);
                if (enums.length > 0) {
                    templateLiteral.push({
                        kind: 'enum',
                        enums,
                    });
                }
            }
        }
        return templateLiteral;
    }
    getEnumOfLiteralType(literalType) {
        if (typescript_1.default.isStringLiteral(literalType.literal)) {
            return {
                type: 'string',
                value: literalType.literal.text
            };
        }
        if (typescript_1.default.isNumericLiteral(literalType.literal)) {
            return {
                type: 'number',
                value: +literalType.literal.text
            };
        }
        if (literalType.literal.kind === typescript_1.default.SyntaxKind.TrueKeyword) {
            return {
                type: 'boolean',
                value: true
            };
        }
        if (literalType.literal.kind === typescript_1.default.SyntaxKind.FalseKeyword) {
            return {
                type: 'boolean',
                value: false
            };
        }
        return {
            type: undefined,
            value: undefined
        };
    }
    getTypeOfLiteralType(literalType, sourceFile) {
        let enumType;
        const enums = [];
        const { type, value } = this.getEnumOfLiteralType(literalType);
        if (type !== undefined) {
            enumType = type;
        }
        if (value !== undefined) {
            enums.push(value);
        }
        if (enumType) {
            return {
                kind: 'enum',
                type: enumType,
                name: enumType,
                enums,
                position: this.getPosition(literalType, sourceFile)
            };
        }
        return {
            kind: undefined,
            position: this.getPosition(literalType, sourceFile)
        };
    }
    getTypeOfUnionTypeOrIntersectionType(unionType, sourceFile) {
        if (unionType.types.every(u => typescript_1.default.isLiteralTypeNode(u))) {
            let enumType;
            const enums = [];
            for (const childType of unionType.types) {
                if (typescript_1.default.isLiteralTypeNode(childType)) {
                    const { type, value } = this.getEnumOfLiteralType(childType);
                    if (type !== undefined) {
                        enumType = type;
                    }
                    if (value !== undefined) {
                        enums.push(value);
                    }
                }
            }
            if (enumType) {
                return {
                    kind: 'enum',
                    type: enumType,
                    name: enumType,
                    enums,
                    position: this.getPosition(unionType, sourceFile)
                };
            }
        }
        else {
            return {
                kind: 'union',
                members: unionType.types.map(u => this.getType(u, sourceFile)),
                position: this.getPosition(unionType, sourceFile)
            };
        }
        return {
            kind: undefined,
            position: this.getPosition(unionType, sourceFile)
        };
    }
    getTypeOfArrayTypeReference(reference, sourceFile) {
        if (reference.typeArguments && reference.typeArguments.length === 1) {
            const typeArgument = reference.typeArguments[0];
            return {
                kind: 'array',
                type: this.getType(typeArgument, sourceFile),
                position: this.getPosition(typeArgument, sourceFile)
            };
        }
        else {
            const position = this.getPosition(reference, sourceFile);
            return {
                kind: 'array',
                type: {
                    kind: undefined,
                    position
                },
                position
            };
        }
    }
    getTypeOfTypeReference(reference, sourceFile) {
        if (typescript_1.default.isIdentifier(reference.typeName)) {
            if (numberTypes.includes(reference.typeName.text)) {
                return {
                    kind: 'number',
                    type: reference.typeName.text,
                    position: this.getPosition(reference.typeName, sourceFile)
                };
            }
            if (reference.typeName.text === 'File') {
                return {
                    kind: 'file',
                    position: this.getPosition(reference.typeName, sourceFile)
                };
            }
            if (reference.typeName.text === 'Array' || reference.typeName.text === 'ReadonlyArray') {
                return this.getTypeOfArrayTypeReference(reference, sourceFile);
            }
            if (reference.typeArguments
                && reference.typeArguments.length > 0) {
                if ((reference.typeName.text === 'Promise'
                    || reference.typeName.text === 'ReturnType'
                    || reference.typeName.text === 'DeepReturnType')) {
                    const typeArgument = reference.typeArguments[0];
                    return this.getType(typeArgument, sourceFile);
                }
                else if (this.checker) {
                    const type = this.getTypeOfComplexType(reference, sourceFile);
                    if (type) {
                        return type;
                    }
                }
                else if (reference.typeName.text === 'Pick') {
                    const type = this.getTypeOfPick(reference.typeName, reference.typeArguments, sourceFile);
                    if (type) {
                        return type;
                    }
                }
            }
            return {
                kind: 'reference',
                name: reference.typeName.text,
                position: this.getPosition(reference.typeName, sourceFile)
            };
        }
        if (typescript_1.default.isQualifiedName(reference.typeName) && typescript_1.default.isIdentifier(reference.typeName.left)) {
            const enumName = reference.typeName.left.text;
            const enumDeclaration = this.declarations.find((m) => m.kind === 'enum' && m.name === enumName);
            if (enumDeclaration) {
                return {
                    kind: 'enum',
                    name: enumDeclaration.name,
                    type: enumDeclaration.type,
                    enums: enumDeclaration.members.map(m => m.value),
                    position: this.getPosition(reference.typeName, sourceFile)
                };
            }
        }
        return {
            kind: undefined,
            position: this.getPosition(reference.typeName, sourceFile)
        };
    }
    getTypeOfTsType(type, typeNode, sourceFile, checker) {
        if (type.flags & typescript_1.default.TypeFlags.String) {
            return {
                kind: 'string',
                position: this.getPosition(typeNode, sourceFile),
            };
        }
        if (type.flags & typescript_1.default.TypeFlags.Number) {
            return {
                kind: 'number',
                type: 'number',
                position: this.getPosition(typeNode, sourceFile),
            };
        }
        if (type.flags & typescript_1.default.TypeFlags.Boolean) {
            return {
                kind: 'boolean',
                position: this.getPosition(typeNode, sourceFile),
            };
        }
        if (type.flags & typescript_1.default.TypeFlags.Void) {
            return {
                kind: 'void',
                position: this.getPosition(typeNode, sourceFile),
            };
        }
        if (type.flags & typescript_1.default.TypeFlags.Null) {
            return {
                kind: 'null',
                position: this.getPosition(typeNode, sourceFile),
            };
        }
        const name = checker.typeToString(type);
        if (name === 'File') {
            return {
                kind: 'file',
                position: this.getPosition(typeNode, sourceFile)
            };
        }
        return {
            kind: 'reference',
            name,
            position: this.getPosition(typeNode, sourceFile),
        };
    }
    getTypeOfPick(typeName, typeArguments, sourceFile) {
        const argument = typeArguments[0];
        if (typescript_1.default.isTypeReferenceNode(argument) && typescript_1.default.isIdentifier(argument.typeName)) {
            const declarationName = argument.typeName.escapedText.toString();
            this.preHandleType(declarationName);
            const declaration = this.declarations.find(m => m.kind === 'object' && m.name === declarationName);
            if (declaration && declaration.kind === 'object') {
                const field = typeArguments[1];
                const memberNames = [];
                if (typescript_1.default.isLiteralTypeNode(field)) {
                    if (typescript_1.default.isStringLiteral(field.literal)) {
                        memberNames.push(field.literal.text);
                    }
                }
                else if (typescript_1.default.isUnionTypeNode(field)) {
                    for (const type of field.types) {
                        if (typescript_1.default.isLiteralTypeNode(type) && typescript_1.default.isStringLiteral(type.literal)) {
                            memberNames.push(type.literal.text);
                        }
                    }
                }
                else {
                    if (!this.disableWarning) {
                        (0, utils_1.warn)(this.getPosition(typeName, sourceFile), 'parser');
                    }
                    return undefined;
                }
                const members = [];
                let minProperties = 0;
                let maxProperties = 0;
                for (const member of declaration.members) {
                    if (memberNames.includes(member.name)) {
                        maxProperties++;
                        if (!member.optional) {
                            minProperties++;
                        }
                        members.push(member);
                    }
                }
                return {
                    kind: 'object',
                    members,
                    minProperties,
                    maxProperties,
                    position: this.getPosition(typeName, sourceFile)
                };
            }
        }
        return undefined;
    }
    getTypeOfComplexType(referenceNode, sourceFile) {
        if (!this.checker) {
            return undefined;
        }
        const type = this.checker.getTypeAtLocation(referenceNode);
        if (type.flags & typescript_1.default.TypeFlags.Object) {
            const members = [];
            let minProperties = 0;
            let maxProperties = 0;
            for (const property of type.getProperties()) {
                const memberType = this.checker.getTypeOfSymbolAtLocation(property, referenceNode);
                maxProperties++;
                if (!(property.flags & typescript_1.default.SymbolFlags.Optional)) {
                    minProperties++;
                }
                members.push({
                    name: property.getName(),
                    type: this.getTypeOfTsType(memberType, referenceNode.typeName, sourceFile, this.checker),
                });
            }
            return {
                kind: 'object',
                members,
                minProperties,
                maxProperties,
                position: this.getPosition(referenceNode.typeName, sourceFile)
            };
        }
        return undefined;
    }
    getTypeOfTypeLiteral(literal, sourceFile) {
        if (literal.members.length === 1 && typescript_1.default.isIndexSignatureDeclaration(literal.members[0])) {
            const member = literal.members[0];
            if (typescript_1.default.isIndexSignatureDeclaration(member) && member.parameters.length === 1) {
                const parameterType = member.parameters[0]?.type;
                if (parameterType && member.type) {
                    return {
                        kind: 'map',
                        key: this.getType(parameterType, sourceFile),
                        value: this.getType(member.type, sourceFile),
                        position: this.getPosition(parameterType, sourceFile)
                    };
                }
            }
        }
        else {
            const { members, minProperties, maxProperties, additionalProperties } = this.getMembersInfo(literal, sourceFile);
            return {
                kind: 'object',
                members,
                minProperties,
                maxProperties: additionalProperties === undefined ? maxProperties : undefined,
                additionalProperties,
                position: this.getPosition(literal, sourceFile)
            };
        }
        return {
            kind: undefined,
            position: this.getPosition(literal, sourceFile)
        };
    }
    getDecorators(sourceFile, decorators) {
        if (!decorators) {
            return undefined;
        }
        const result = [];
        for (const decorator of decorators) {
            if (typescript_1.default.isCallExpression(decorator.expression) && typescript_1.default.isIdentifier(decorator.expression.expression)) {
                result.push({
                    name: decorator.expression.expression.escapedText.toString(),
                    parameters: decorator.expression.arguments.map((a) => this.getTypeAndValueOfExpression(a, sourceFile).value)
                });
            }
        }
        return result;
    }
    getMembersInfo(node, sourceFile) {
        if (typescript_1.default.isTypeLiteralNode(node)) {
            return this.getObjectMembers(node.members, sourceFile);
        }
        if (typescript_1.default.isUnionTypeNode(node)) {
            return this.getMembersInfoOfUnionType(node, sourceFile);
        }
        if (typescript_1.default.isIntersectionTypeNode(node)) {
            return this.getMembersInfoOfIntersectionType(node, sourceFile);
        }
        if (typescript_1.default.isParenthesizedTypeNode(node)) {
            return this.getMembersInfoOfParenthesizedType(node, sourceFile);
        }
        if (typescript_1.default.isTypeReferenceNode(node)) {
            return this.getMembersInfoOfTypeReference(node);
        }
        return { members: [], minProperties: 0, maxProperties: 0 };
    }
    getMembersInfoOfUnionType(unionType, sourceFile) {
        const members = [];
        let minProperties = Infinity;
        let maxProperties = 0;
        for (const type of unionType.types) {
            const childMembersInfo = this.getMembersInfo(type, sourceFile);
            if (minProperties > childMembersInfo.minProperties) {
                minProperties = childMembersInfo.minProperties;
            }
            if (maxProperties < childMembersInfo.maxProperties) {
                maxProperties = childMembersInfo.maxProperties;
            }
            if (members.length === 0) {
                const childMembers = JSON.parse(JSON.stringify(childMembersInfo.members));
                members.push(...childMembers);
            }
            else {
                const childMembers = childMembersInfo.members;
                this.setOptionalAndEnumOfUnionType(members, childMembers);
                for (const member of childMembers) {
                    if (members.every(m => m.name !== member.name)) {
                        const childMember = JSON.parse(JSON.stringify(member));
                        childMember.optional = true;
                        members.push(childMember);
                    }
                }
            }
        }
        return { members, minProperties, maxProperties };
    }
    setOptionalAndEnumOfUnionType(members, childMembers) {
        for (const member of members) {
            const childMember = childMembers.find(m => m.name === member.name);
            if (!childMember) {
                member.optional = true;
            }
            else {
                this.setEnumOfUnionType(member, childMember);
            }
        }
    }
    setEnumOfUnionType(member, childMember) {
        if (childMember.type.kind === 'enum' && member.type.kind === 'enum') {
            for (const typeEnum of childMember.type.enums) {
                if (member.type.enums.every(e => e !== typeEnum)) {
                    member.type.enums.push(typeEnum);
                }
            }
        }
    }
    getMembersInfoOfIntersectionType(intersectionType, sourceFile) {
        const members = [];
        let minProperties = 0;
        let maxProperties = 0;
        for (const type of intersectionType.types) {
            const childMembersInfo = this.getMembersInfo(type, sourceFile);
            minProperties += childMembersInfo.minProperties;
            maxProperties += childMembersInfo.maxProperties;
            const childMembers = childMembersInfo.members;
            for (const member of childMembers) {
                if (members.every(m => m.name !== member.name)) {
                    members.push(JSON.parse(JSON.stringify(member)));
                }
            }
        }
        return { members, minProperties, maxProperties };
    }
    getMembersInfoOfParenthesizedType(parenthesizedType, sourceFile) {
        const members = [];
        const childMembersInfo = this.getMembersInfo(parenthesizedType.type, sourceFile);
        const minProperties = childMembersInfo.minProperties;
        const maxProperties = childMembersInfo.maxProperties;
        const childMembers = JSON.parse(JSON.stringify(childMembersInfo.members));
        for (const member of childMembers) {
            members.push(member);
        }
        return { members, minProperties, maxProperties };
    }
    getMembersInfoOfTypeReference(node) {
        const members = [];
        let minProperties = 0;
        let maxProperties = 0;
        const referenceName = node.typeName.text;
        this.preHandleType(referenceName);
        const objectDeclaration = this.declarations.find(m => m.kind === 'object' && m.name === referenceName);
        if (objectDeclaration && objectDeclaration.kind === 'object') {
            for (const member of objectDeclaration.members) {
                if (members.every(m => m.name !== member.name)) {
                    members.push(JSON.parse(JSON.stringify(member)));
                    maxProperties++;
                    if (!member.optional) {
                        minProperties++;
                    }
                }
            }
        }
        return { members, minProperties, maxProperties };
    }
    preHandleType(typeName) {
        // if the node is pre-handled, then it should be in `declarations` already, so don't continue
        if (this.declarations.some(m => m.name === typeName)) {
            return;
        }
        let findIt = false;
        for (const sourceFile of this.sourceFiles) {
            if (findIt) {
                return;
            }
            typescript_1.default.forEachChild(sourceFile, node => {
                if (findIt) {
                    return;
                }
                if (typescript_1.default.isInterfaceDeclaration(node)) {
                    if (node.name.text === typeName) {
                        findIt = true;
                        this.handleSourceFile(node, sourceFile);
                    }
                }
                else if (typescript_1.default.isTypeAliasDeclaration(node) && node.name.text === typeName) {
                    findIt = true;
                    this.handleSourceFile(node, sourceFile);
                }
            });
        }
    }
    getObjectMembers(elements, sourceFile) {
        const members = [];
        let minProperties = 0;
        let maxProperties = 0;
        let additionalProperties;
        for (const element of elements) {
            if (typescript_1.default.isPropertySignature(element) || typescript_1.default.isPropertyDeclaration(element)) {
                const member = this.getObjectMemberOfPropertyOrMethodOrConstructorParameterOrCallSignature(element, sourceFile);
                members.push(member);
                if (!element.questionToken) {
                    minProperties++;
                }
                maxProperties++;
            }
            else if (typescript_1.default.isIndexSignatureDeclaration(element)) {
                if (element.type) {
                    additionalProperties = this.getType(element.type, sourceFile);
                }
            }
            else if (typescript_1.default.isMethodSignature(element) || typescript_1.default.isMethodDeclaration(element) || typescript_1.default.isCallSignatureDeclaration(element)) {
                const member = this.getObjectMemberOfPropertyOrMethodOrConstructorParameterOrCallSignature(element, sourceFile);
                members.push(member);
            }
            else if (typescript_1.default.isConstructorDeclaration(element)) {
                for (const parameter of element.parameters) {
                    if (parameter.modifiers
                        && parameter.modifiers.some((modifier) => modifier.kind === typescript_1.default.SyntaxKind.PublicKeyword
                            || modifier.kind === typescript_1.default.SyntaxKind.PrivateKeyword
                            || modifier.kind === typescript_1.default.SyntaxKind.ProtectedKeyword)) {
                        const member = this.getObjectMemberOfPropertyOrMethodOrConstructorParameterOrCallSignature(parameter, sourceFile);
                        members.push(member);
                        if (!parameter.questionToken) {
                            minProperties++;
                        }
                        maxProperties++;
                    }
                }
            }
        }
        return { members, minProperties, maxProperties, additionalProperties };
    }
    getExpression(expression, name, sourceFile) {
        const { type, value } = this.getTypeAndValueOfExpression(expression, sourceFile);
        return {
            name: name.text,
            type,
            value
        };
    }
    getTypeAndValueOfExpression(expression, sourceFile) {
        if (typescript_1.default.isStringLiteral(expression)) {
            return {
                type: {
                    kind: 'string',
                    position: this.getPosition(expression, sourceFile)
                },
                value: expression.text
            };
        }
        if (typescript_1.default.isNumericLiteral(expression)) {
            return {
                type: {
                    kind: 'number',
                    type: 'number',
                    position: this.getPosition(expression, sourceFile)
                },
                value: +expression.text
            };
        }
        if (expression.kind === typescript_1.default.SyntaxKind.FalseKeyword || expression.kind === typescript_1.default.SyntaxKind.TrueKeyword) {
            return {
                type: {
                    kind: 'boolean',
                    position: this.getPosition(expression, sourceFile)
                },
                value: expression.kind === typescript_1.default.SyntaxKind.TrueKeyword
            };
        }
        else if (typescript_1.default.isArrayLiteralExpression(expression)) {
            let elementsType = {
                kind: undefined,
                position: this.getPosition(expression, sourceFile)
            };
            const elementsValues = [];
            for (const element of expression.elements) {
                const { type, value } = this.getTypeAndValueOfExpression(element, sourceFile);
                elementsType = type;
                elementsValues.push(value);
            }
            return {
                type: {
                    kind: 'array',
                    type: elementsType,
                    position: this.getPosition(expression, sourceFile)
                },
                value: elementsValues
            };
        }
        else if (typescript_1.default.isObjectLiteralExpression(expression)) {
            const members = [];
            const value = {};
            for (const property of expression.properties) {
                if (typescript_1.default.isPropertyAssignment(property) && typescript_1.default.isIdentifier(property.name)) {
                    const expression = this.getExpression(property.initializer, property.name, sourceFile);
                    members.push({
                        name: expression.name,
                        type: expression.type
                    });
                    value[expression.name] = expression.value;
                }
            }
            return {
                type: {
                    kind: 'object',
                    members,
                    minProperties: members.length,
                    position: this.getPosition(expression, sourceFile)
                },
                value
            };
        }
        else if (typescript_1.default.isIdentifier(expression)) {
            return {
                type: {
                    kind: 'reference',
                    name: expression.escapedText.toString(),
                    position: this.getPosition(expression, sourceFile)
                },
                value: expression.escapedText.toString(),
            };
        }
        return {
            type: {
                kind: undefined,
                position: this.getPosition(expression, sourceFile)
            },
            value: undefined
        };
    }
    getObjectMemberOfPropertyOrMethodOrConstructorParameterOrCallSignature(property, sourceFile) {
        const member = {
            name: !typescript_1.default.isCallSignatureDeclaration(property) && typescript_1.default.isIdentifier(property.name) ? property.name.text : '',
            type: {
                kind: undefined,
                position: this.getPosition(property, sourceFile)
            },
            decorators: this.getDecorators(sourceFile, property.decorators),
        };
        if (property.questionToken) {
            member.optional = true;
        }
        let defaultValue;
        if ((typescript_1.default.isPropertySignature(property) || typescript_1.default.isPropertyDeclaration(property))
            && property.initializer) {
            const { type, value } = this.getTypeAndValueOfExpression(property.initializer, sourceFile);
            member.type = type;
            defaultValue = value;
        }
        if (property.type) {
            member.type = this.getType(property.type, sourceFile);
        }
        if (defaultValue !== undefined) {
            member.type.default = defaultValue;
        }
        if (typescript_1.default.isPropertySignature(property) || typescript_1.default.isPropertyDeclaration(property)) {
            this.setPropertyJsDoc(property, member, sourceFile);
        }
        if (typescript_1.default.isMethodSignature(property) || typescript_1.default.isMethodDeclaration(property) || typescript_1.default.isCallSignatureDeclaration(property)) {
            const { jsDocs, comments } = this.getJsDocs(property, sourceFile);
            member.comments = comments;
            member.jsDocs = jsDocs;
            member.parameters = property.parameters.map((parameter) => this.handleFunctionParameter(parameter, sourceFile));
        }
        return member;
    }
    getParameter(parameter, sourceFile) {
        let type;
        let value;
        if (parameter.initializer) {
            const typeAndValue = this.getTypeAndValueOfExpression(parameter.initializer, sourceFile);
            type = typeAndValue.type;
            value = typeAndValue.value;
        }
        if (parameter.type) {
            type = this.getType(parameter.type, sourceFile);
        }
        if (!type) {
            type = {
                kind: undefined,
                position: this.getPosition(parameter, sourceFile)
            };
        }
        type.default = value;
        return {
            name: parameter.name.text,
            type,
            optional: !!parameter.questionToken || type.default !== undefined,
            decorators: this.getDecorators(sourceFile, parameter.decorators),
        };
    }
    setJsDoc(jsDoc, type) {
        if (jsDoc.name === 'mapValueType') {
            this.setJsDocMapValue(jsDoc, type);
        }
        else if (jsDoc.name === 'type') {
            this.overrideType(type, jsDoc);
        }
        else if (type.kind === 'array') {
            this.setJsDocArray(jsDoc, type);
        }
        else if (type.kind === 'number') {
            this.setJsDocNumber(jsDoc, type);
        }
        else if (type.kind === 'string') {
            this.setJsDocString(jsDoc, type);
        }
        else if (type.kind === 'boolean') {
            this.setJsDocBoolean(jsDoc, type);
        }
        else if (type.kind === 'object') {
            this.setJsDocObject(jsDoc, type);
        }
        else if (type.kind === 'reference') {
            this.setJsDocReference(jsDoc, type);
        }
        else if (type.kind === 'enum') {
            this.setJsDocEnum(jsDoc, type);
        }
    }
    setPropertyJsDoc(property, member, sourceFile) {
        const { jsDocs, comments } = this.getJsDocs(property, sourceFile);
        member.comments = comments;
        member.jsDocs = jsDocs;
        for (const propertyJsDoc of jsDocs || []) {
            if (propertyJsDoc.name === 'tag') {
                this.setJsDocTag(propertyJsDoc, member);
            }
            else if (propertyJsDoc.name === 'param') {
                this.setJsDocParam(propertyJsDoc, member);
            }
            else if (propertyJsDoc.name === 'index') {
                member.index = true;
            }
            else if (propertyJsDoc.name === 'unique') {
                member.unique = true;
            }
            else if (propertyJsDoc.name === 'sparse') {
                member.sparse = true;
            }
            else if (propertyJsDoc.name === 'select') {
                member.select = true;
            }
            else if (propertyJsDoc.name === 'alias') {
                if (propertyJsDoc.comment) {
                    member.alias = propertyJsDoc.comment;
                }
            }
            else {
                this.setJsDoc(propertyJsDoc, member.type);
            }
        }
    }
    setJsDocReference(propertyJsDoc, type) {
        if (propertyJsDoc.comment) {
            try {
                type.default = JSON.parse(this.getJsDocComment(propertyJsDoc.comment));
            }
            catch (error) {
                if (!this.disableWarning) {
                    (0, utils_1.warn)(type.position, 'parser');
                }
            }
        }
    }
    setJsDocTag(propertyJsDoc, member) {
        if (propertyJsDoc.comment) {
            member.tag = +propertyJsDoc.comment;
        }
    }
    setJsDocMapValue(propertyJsDoc, type) {
        if (propertyJsDoc.comment && type.kind === 'map' && type.value.kind === 'number') {
            type.value.type = propertyJsDoc.comment;
        }
    }
    setJsDocParam(propertyJsDoc, member) {
        if (propertyJsDoc.paramName && propertyJsDoc.type) {
            if (!member.parameters) {
                member.parameters = [];
            }
            member.parameters.push({
                name: propertyJsDoc.paramName,
                type: propertyJsDoc.type,
                optional: propertyJsDoc.optional
            });
        }
        this.overrideType(member.type, propertyJsDoc);
    }
    setJsDocBoolean(propertyJsDoc, type) {
        if (propertyJsDoc.comment) {
            if (propertyJsDoc.name === 'default') {
                type.default = this.getJsDocComment(propertyJsDoc.comment).toLowerCase() === 'true';
            }
            else if (propertyJsDoc.name === 'title') {
                type.title = propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'description') {
                type.description = propertyJsDoc.comment;
            }
        }
    }
    setJsDocEnum(propertyJsDoc, type) {
        if (propertyJsDoc.comment) {
            if (propertyJsDoc.name === 'default') {
                type.default = this.getJsDocComment(propertyJsDoc.comment);
            }
        }
    }
    getJsDocComment(comment) {
        if (comment.length >= 2
            && ((comment.startsWith(`'`) && comment.endsWith(`'`))
                || (comment.startsWith(`"`) && comment.endsWith(`"`))
                || (comment.startsWith('`') && comment.endsWith('`')))) {
            return comment.substring(1, comment.length - 1);
        }
        return comment;
    }
    setJsDocString(propertyJsDoc, type) {
        if (propertyJsDoc.comment) {
            if (propertyJsDoc.name === 'minLength') {
                type.minLength = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'maxLength') {
                type.maxLength = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'pattern') {
                type.pattern = propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'default') {
                type.default = this.getJsDocComment(propertyJsDoc.comment);
            }
            else if (propertyJsDoc.name === 'title') {
                type.title = propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'description') {
                type.description = propertyJsDoc.comment;
            }
        }
        if (propertyJsDoc.name === 'lowercase') {
            type.lowercase = true;
        }
        else if (propertyJsDoc.name === 'uppercase') {
            type.uppercase = true;
        }
        else if (propertyJsDoc.name === 'trim') {
            type.trim = true;
        }
    }
    setJsDocNumber(propertyJsDoc, type) {
        if (propertyJsDoc.comment) {
            if (propertyJsDoc.name === 'multipleOf') {
                type.multipleOf = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'maximum') {
                type.maximum = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'minimum') {
                type.minimum = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'exclusiveMaximum') {
                type.exclusiveMaximum = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'exclusiveMinimum') {
                type.exclusiveMinimum = +propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'default') {
                type.default = +this.getJsDocComment(propertyJsDoc.comment);
            }
            else if (propertyJsDoc.name === 'title') {
                type.title = propertyJsDoc.comment;
            }
            else if (propertyJsDoc.name === 'description') {
                type.description = propertyJsDoc.comment;
            }
        }
    }
    overrideType(type, jsDoc) {
        if (jsDoc && jsDoc.comment) {
            if (type.kind === 'number') {
                type.type = jsDoc.comment;
            }
            else if (type.kind === 'array' && type.type.kind === 'number') {
                type.type = {
                    kind: type.type.kind,
                    type: jsDoc.comment,
                    position: type.type.position
                };
            }
        }
    }
    setJsDocArray(jsDoc, type) {
        if (jsDoc.comment) {
            if (jsDoc.name === 'minItems') {
                type.minItems = +jsDoc.comment;
            }
            else if (jsDoc.name === 'maxItems') {
                type.maxItems = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemType') {
                this.overrideType(type, jsDoc);
            }
            else if (type.type.kind === 'number') {
                this.setJsDocNumberArray(jsDoc, type.type);
            }
            else if (type.type.kind === 'string') {
                this.setJsDocStringArray(jsDoc, type.type);
            }
            else if (type.type.kind === 'boolean') {
                if (jsDoc.name === 'itemDefault') {
                    type.type.default = this.getJsDocComment(jsDoc.comment).toLowerCase() === 'true';
                }
            }
            else if (jsDoc.name === 'default') {
                type.default = JSON.parse(this.getJsDocComment(jsDoc.comment));
            }
            else if (jsDoc.name === 'title') {
                type.title = jsDoc.comment;
            }
            else if (jsDoc.name === 'description') {
                type.description = jsDoc.comment;
            }
        }
        else if (jsDoc.name === 'uniqueItems') {
            type.uniqueItems = true;
        }
    }
    setJsDocNumberArray(jsDoc, type) {
        if (jsDoc.comment) {
            if (jsDoc.name === 'itemMultipleOf') {
                type.multipleOf = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemMinimum') {
                type.minimum = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemMaximum') {
                type.maximum = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemExclusiveMinimum') {
                type.exclusiveMinimum = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemExclusiveMaximum') {
                type.exclusiveMaximum = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemDefault') {
                type.default = +this.getJsDocComment(jsDoc.comment);
            }
        }
    }
    setJsDocStringArray(jsDoc, type) {
        if (jsDoc.comment) {
            if (jsDoc.name === 'itemMinLength') {
                type.minLength = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemMaxLength') {
                type.maxLength = +jsDoc.comment;
            }
            else if (jsDoc.name === 'itemPattern') {
                type.pattern = jsDoc.comment;
            }
            else if (jsDoc.name === 'itemDefault') {
                type.default = this.getJsDocComment(jsDoc.comment);
            }
        }
    }
    setJsDocObject(jsDoc, type) {
        if (jsDoc.comment) {
            if (jsDoc.name === 'minProperties') {
                type.minProperties = +jsDoc.comment;
            }
            else if (jsDoc.name === 'maxProperties') {
                type.maxProperties = +jsDoc.comment;
            }
            else if (jsDoc.name === 'default') {
                type.default = JSON.parse(this.getJsDocComment(jsDoc.comment));
            }
            else if (jsDoc.name === 'title') {
                type.title = jsDoc.comment;
            }
            else if (jsDoc.name === 'description') {
                type.description = jsDoc.comment;
            }
        }
        else {
            if (jsDoc.name === 'additionalProperties') {
                type.additionalProperties = true;
            }
        }
    }
    getPosition(typeNode, sourceFile) {
        return {
            file: this.getRelativePath(sourceFile.fileName),
            ...typescript_1.default.getLineAndCharacterOfPosition(sourceFile, typeNode.getStart(sourceFile))
        };
    }
    getModifiers(typeNode) {
        const modifiers = [];
        if (typeNode.modifiers) {
            for (const modifier of typeNode.modifiers) {
                if (modifier.kind === typescript_1.default.SyntaxKind.AbstractKeyword) {
                    modifiers.push('abstract');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.AsyncKeyword) {
                    modifiers.push('async');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.ConstKeyword) {
                    modifiers.push('const');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.DeclareKeyword) {
                    modifiers.push('declare');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.DefaultKeyword) {
                    modifiers.push('default');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.ExportKeyword) {
                    modifiers.push('export');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.PrivateKeyword) {
                    modifiers.push('private');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.ProtectedKeyword) {
                    modifiers.push('protected');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.PublicKeyword) {
                    modifiers.push('public');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.OverrideKeyword) {
                    modifiers.push('override');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.ReadonlyKeyword) {
                    modifiers.push('readonly');
                }
                if (modifier.kind === typescript_1.default.SyntaxKind.StaticKeyword) {
                    modifiers.push('static');
                }
            }
        }
        return modifiers.length > 0 ? modifiers : undefined;
    }
}
exports.Parser = Parser;
const numberTypes = ['double', 'float', 'uint32', 'fixed32', 'integer', 'int32', 'sint32', 'sfixed32', 'uint64', 'fixed64', 'int64', 'sint64', 'sfixed64'];
