"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateTypescriptOfType = exports.generateTypescriptOfFunctionParameter = exports.generateTypescriptOfFunctionDeclaration = exports.generateTypescriptOfEnumMember = exports.generateTypescriptOfEnumDeclaration = exports.generateTypescriptOfObjectAdditionalProperties = exports.generateTypescriptOfObjectDeclaration = exports.generateTypescriptOfUnionDeclaration = exports.generateTypescriptOfStringOrNumberDeclaration = exports.generateTypescriptOfReferenceDeclaration = exports.generateTypescript = void 0;
/**
 * @public
 */
function generateTypescript(declarations) {
    const messages = [];
    for (const declaration of declarations) {
        if (declaration.kind === 'function') {
            if (declaration.name) {
                messages.push(generateTypescriptOfFunctionDeclaration(declaration));
            }
        }
        else if (declaration.kind === 'enum') {
            messages.push(generateTypescriptOfEnumDeclaration(declaration));
        }
        else if (declaration.kind === 'object') {
            messages.push(generateTypescriptOfObjectDeclaration(declaration));
        }
        else if (declaration.kind === 'union') {
            messages.push(generateTypescriptOfUnionDeclaration(declaration));
        }
        else if (declaration.kind === 'string' || declaration.kind === 'number') {
            messages.push(generateTypescriptOfStringOrNumberDeclaration(declaration));
        }
        else if (declaration.kind === 'reference') {
            messages.push(generateTypescriptOfReferenceDeclaration(declaration));
        }
    }
    return `/* eslint-disable */

${messages.join('\n\n')}
`;
}
exports.generateTypescript = generateTypescript;
/**
 * @public
 */
function generateTypescriptOfReferenceDeclaration(declaration) {
    const comments = declaration.comments ? declaration.comments.join('\n') + '\n' : '';
    return `${comments}type ${declaration.newName} = ${declaration.name}`;
}
exports.generateTypescriptOfReferenceDeclaration = generateTypescriptOfReferenceDeclaration;
/**
 * @public
 */
function generateTypescriptOfStringOrNumberDeclaration(declaration) {
    const comments = declaration.comments ? declaration.comments.join('\n') + '\n' : '';
    const members = declaration.enums ? declaration.enums.map((e) => JSON.stringify(e)).join(' | ') : declaration.kind;
    return `${comments}type ${declaration.name} = ${members}`;
}
exports.generateTypescriptOfStringOrNumberDeclaration = generateTypescriptOfStringOrNumberDeclaration;
/**
 * @public
 */
function generateTypescriptOfUnionDeclaration(declaration) {
    const members = declaration.members.map((m) => generateTypescriptOfType(m));
    const comments = declaration.comments ? declaration.comments.join('\n') + '\n' : '';
    return `${comments}type ${declaration.name} = ${members.join(' | ')}`;
}
exports.generateTypescriptOfUnionDeclaration = generateTypescriptOfUnionDeclaration;
/**
 * @public
 */
function generateTypescriptOfObjectDeclaration(declaration) {
    const members = declaration.members.map((m) => {
        const comments = m.comments ? m.comments.join('\n') + '\n  ' : '';
        return '  ' + comments + generateTypescriptOfMember(m);
    });
    if (declaration.additionalProperties) {
        members.push('  ' + generateTypescriptOfObjectAdditionalProperties(declaration.additionalProperties));
    }
    const comments = declaration.comments ? declaration.comments.join('\n') + '\n' : '';
    return `${comments}interface ${declaration.name} {
${members.join('\n')}
}`;
}
exports.generateTypescriptOfObjectDeclaration = generateTypescriptOfObjectDeclaration;
function generateTypescriptOfMember(member, processChild) {
    if (member.parameters) {
        const parameters = member.parameters.map((m) => generateTypescriptOfFunctionParameter(m));
        return `${member.name}(${parameters.join(', ')}): ${generateTypescriptOfType(member.type, processChild)}`;
    }
    return generateTypescriptOfFunctionParameter(member, processChild);
}
/**
 * @public
 */
function generateTypescriptOfObjectAdditionalProperties(additionalProperties) {
    const type = additionalProperties === true ? 'unknown' : generateTypescriptOfType(additionalProperties);
    return `[name: string]: ${type}`;
}
exports.generateTypescriptOfObjectAdditionalProperties = generateTypescriptOfObjectAdditionalProperties;
/**
 * @public
 */
function generateTypescriptOfEnumDeclaration(declaration) {
    const members = declaration.members.map(generateTypescriptOfEnumMember);
    return `const enum ${declaration.name} {
${members.join('\n')}
}`;
}
exports.generateTypescriptOfEnumDeclaration = generateTypescriptOfEnumDeclaration;
/**
 * @public
 */
function generateTypescriptOfEnumMember(member) {
    const value = JSON.stringify(member.value);
    return `  ${member.name} = ${value},`;
}
exports.generateTypescriptOfEnumMember = generateTypescriptOfEnumMember;
/**
 * @public
 */
function generateTypescriptOfFunctionDeclaration(declaration) {
    const parameters = declaration.parameters.map((m) => generateTypescriptOfFunctionParameter(m));
    const type = generateTypescriptOfType(declaration.type);
    const comments = declaration.comments ? declaration.comments.join('\n') + '\n' : '';
    return `${comments}declare function ${declaration.name}(${parameters.join(', ')}): ${type}`;
}
exports.generateTypescriptOfFunctionDeclaration = generateTypescriptOfFunctionDeclaration;
/**
 * @public
 */
function generateTypescriptOfFunctionParameter(parameter, processChild) {
    const optional = parameter.optional ? '?' : '';
    return `${parameter.name}${optional}: ${generateTypescriptOfType(parameter.type, processChild)}`;
}
exports.generateTypescriptOfFunctionParameter = generateTypescriptOfFunctionParameter;
function generateTypescriptTemplateLiteral(templateLiteral) {
    let result = '';
    for (const part of templateLiteral) {
        if (part.kind === 'string') {
            result += '${string}';
        }
        else if (part.kind === 'number') {
            result += '${number}';
        }
        else if (part.kind === 'boolean') {
            result += '${boolean}';
        }
        else if (part.kind === 'enum') {
            if (part.enums.length === 1) {
                result += part.enums[0];
            }
            else {
                result += `\${${part.enums.map((e) => JSON.stringify(e)).join(' | ')}}`;
            }
        }
    }
    return `\`${result}\``;
}
/**
 * @public
 */
function generateTypescriptOfType(type, processChild) {
    if (processChild) {
        const childResult = processChild(type);
        if (childResult !== undefined) {
            return childResult;
        }
    }
    if (type.kind === 'enum') {
        return type.enums.map((e) => JSON.stringify(e)).join(' | ');
    }
    if (type.kind === 'string' && type.templateLiteral) {
        return generateTypescriptTemplateLiteral(type.templateLiteral);
    }
    if (type.kind === 'boolean' || type.kind === 'number' || type.kind === 'string' || type.kind === 'null') {
        return type.kind;
    }
    if (type.kind === 'array') {
        const item = generateTypescriptOfType(type.type, processChild);
        if (type.minItems && type.maxItems && type.minItems === type.maxItems) {
            return `[${Array.from({ length: type.minItems }).fill(item).join(', ')}]`;
        }
        if ((type.type.kind === 'enum' || type.type.kind === 'number' || type.type.kind === 'string') && type.type.enums && type.type.enums.length > 1) {
            return `(${item})[]`;
        }
        return `${item}[]`;
    }
    if (type.kind === 'reference') {
        return type.name;
    }
    if (type.kind === 'file') {
        return 'File';
    }
    if (type.kind === 'void') {
        return 'void';
    }
    if (type.kind === 'union') {
        return type.members.map((e) => generateTypescriptOfType(e, processChild)).join(' | ');
    }
    if (type.kind === 'object') {
        const members = type.members.map((m) => generateTypescriptOfMember(m, processChild));
        if (type.additionalProperties) {
            members.push(generateTypescriptOfObjectAdditionalProperties(type.additionalProperties));
        }
        return `{ ${members.join(', ')} }`;
    }
    if (type.kind === 'map') {
        const keyType = generateTypescriptOfType(type.key, processChild);
        const valueType = generateTypescriptOfType(type.value, processChild);
        return `{ [name: ${keyType}]: ${valueType} }`;
    }
    return 'unknown';
}
exports.generateTypescriptOfType = generateTypescriptOfType;
